/******************************************************************************
 *   Copyright (C) 2008 by Niko Valimaki <nvalimak@cs.helsinki.fi>            *
 *   Text collection interface for an in-memory XQuery/XPath engine           *
 *                                                                            *
 *   This program is free software; you can redistribute it and/or modify     *
 *   it under the terms of the GNU Lesser General Public License as published *
 *   by the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                      *
 *                                                                            *
 *   This program is distributed in the hope that it will be useful,          *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *   GNU Lesser General Public License for more details.                      *
 *                                                                            *
 *   You should have received a copy of the GNU Lesser General Public License *
 *   along with this program; if not, write to the                            *
 *   Free Software Foundation, Inc.,                                          *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                *
 ******************************************************************************/ 

#ifndef _SXSI_TextCollection_h_
#define _SXSI_TextCollection_h_

#include "Tools.h" // Defines ulong and uchar.
#include <vector>
#include <utility> // Defines std::pair.

namespace SXSI
{

    /**
     * General interface for a text collection
     *
     * Class is virtual, make objects by calling 
     * the static method InitTextCollection().
     */
    class TextCollection
    {
    public:
        // Type of document identifier
        typedef int DocId;
        // Type for text position (FIXME ulong or long?)
        typedef ulong TextPosition;

        // mode flag: Default includes both the index and "naive" text
        enum index_mode_t { index_mode_default, index_mode_text_only }; // index_mode_index_only

        /**
         * Load from a file
         *
         * The second parameter is a prefix to be used for multiple
         * files. (SWCSAWrapper uses multiple save files!)
         *
         * New samplerate can be given, otherwise will use the one specified in the save file!
         *
         * Throws an exception if std::fread() fails.
         * 
         */
        static TextCollection* Load(FILE *, char const *, index_mode_t = index_mode_default, unsigned samplerate = 0);

        /**
         * Save data structure into a file
         *
         * The second parameter is a prefix to be used for multiple
         * files. (SWCSAWrapper uses multiple save files!)
         *
         * Throws an exception if std::fwrite() fails.
         */
        virtual void Save(FILE *, char const *) const = 0;

        /**
         * Virtual destructor
         */
        virtual ~TextCollection() { };

	/**
	 * Tests if the string pointed to by DocId is empty
         */
	virtual bool EmptyText(DocId) const = 0;

        /**
         * Displaying content
         *
         * Returns the i'th text in the collection.
         * The numbering starts from 0.
         *
         * Call DeleteText() for each pointer returned by GetText()
         * to avoid possible memory leaks.
         */
        virtual uchar* GetText(DocId) const = 0;
        virtual void DeleteText(uchar *text) const = 0;

        /**
         * Returns a pointer to the beginning of texts i, i+1, ..., j.
         * Texts are separated by a '\0' byte.
         *
         * Call DeleteText() for each pointer returned by GetText()
         * to avoid possible memory leaks.
         */
        virtual uchar * GetText(DocId i, DocId j) const = 0;

        /**
         * Returns substring [i, j] of k'th text
         *
         * Note: Parameters i and j are text positions inside the k'th text.
         */
//        virtual uchar* GetText(DocId, TextPosition, TextPosition) const = 0;
        /**
         * Returns backwards (reverse) iterator to the end of i'th text
         * 
         * Note: Do we need this?
         * Forward iterator would be really in-efficient compared to
         * getText(k).
         *
         * TODO Define and implement const_reverse_iterator.
         */
        //const_reverse_iterator rend(DocId) const;
        
        /**
         * Existential queries
         */
        // Is there a text prefixed by given string?
        virtual bool IsPrefix(uchar const *) const = 0;
        // Is there a text having given string as a suffix?
        virtual bool IsSuffix(uchar const *) const = 0;
        // Is there a text that equals given string?
        virtual bool IsEqual(uchar const *) const = 0;
        // Does a text contain given string?
        virtual bool IsContains(uchar const *) const = 0;
        // Is there a text that is lexicographically less than given string?
        virtual bool IsLessThan(uchar const *) const = 0;

        /**
         * Existential queries for given DocId interval.
         */
        virtual bool IsPrefix(uchar const *, DocId, DocId) const = 0;
        virtual bool IsSuffix(uchar const *, DocId, DocId) const = 0;
        virtual bool IsEqual(uchar const *, DocId, DocId) const = 0;
        virtual bool IsContains(uchar const *, DocId, DocId) const = 0;
        virtual bool IsLessThan(uchar const *, DocId, DocId) const = 0;

        /**
         * Counting queries
         * Result is the number of occurrences.
         */
        virtual ulong Count(uchar const *) const = 0;
        /**
         * More counting queries
         * Result is the number of documents.
         */
        virtual unsigned CountPrefix(uchar const *) const = 0;
        virtual unsigned CountSuffix(uchar const *) const = 0;
        virtual unsigned CountEqual(uchar const *) const = 0;
        virtual unsigned CountContains(uchar const *) const = 0;
        virtual unsigned CountLessThan(uchar const *) const = 0;

        /**
         * Counting queries for given DocId interval
         */
        virtual unsigned CountPrefix(uchar const *, DocId, DocId) const = 0;
        virtual unsigned CountSuffix(uchar const *, DocId, DocId) const = 0;
        virtual unsigned CountEqual(uchar const *, DocId, DocId) const = 0;
        virtual unsigned CountContains(uchar const *, DocId, DocId) const = 0;
        virtual unsigned CountLessThan(uchar const *, DocId, DocId) const = 0;

        /**
         * Document reporting queries
         *
         * Result is a vector of document id's in some undefined order.
         */
        // Data type for results
        typedef std::vector<DocId> document_result;
        virtual document_result Prefix(uchar const *) const = 0;
        virtual document_result Suffix(uchar const *) const = 0;
        virtual document_result Equal(uchar const *) const = 0;
        virtual document_result Contains(uchar const *) const = 0;
        virtual document_result LessThan(uchar const *) const = 0;
        virtual document_result KMismaches(uchar const *, unsigned) const = 0;
        virtual document_result KErrors(uchar const *, unsigned) const = 0;
    
        /**
         * Document reporting queries for given DocId interval.
         */
        virtual document_result Prefix(uchar const *, DocId, DocId) const = 0;
        virtual document_result Suffix(uchar const *, DocId, DocId) const = 0;
        virtual document_result Equal(uchar const *, DocId, DocId) const = 0;
        virtual document_result Contains(uchar const *, DocId, DocId) const = 0;
        virtual document_result LessThan(uchar const *, DocId, DocId) const = 0;

        /**
         * Full reporting queries
         *
         * Result is a vector of pairs <doc id, offset> in some undefined order.
         */
        // Data type for results
        typedef std::vector<std::pair<DocId, TextPosition> > full_result;
        virtual full_result FullContains(uchar const *) const = 0;
        // Full reporting query for given DocId interval
        virtual full_result FullContains(uchar const *, DocId, DocId) const = 0;

        virtual full_result FullKMismatches(uchar const *, unsigned) const = 0;
        virtual full_result FullKErrors(uchar const *, unsigned) const = 0;


        virtual TextPosition getLength() const
        {
            std::cerr << "TextCollection::getLength() is unsupported! Use RLCSA instead." << std::endl;
            std::exit(2);
            return 0;
        }
    
        virtual TextPosition LF(uchar c, TextPosition i) const
        {
            std::cerr << "TextCollection::LF() is unsupported! Use RLCSA instead." << std::endl;
            std::exit(2);
            return 0;
        }

        virtual uchar* getSuffix(TextPosition pos, unsigned l) const
        {
            std::cerr << "TextCollection::getSuffix() is unsupported! Use RLCSA instead." << std::endl;
            std::exit(2);
            return 0;
        }
    
        virtual DocId getDoc(TextPosition i) const
        {
            std::cerr << "TextCollection::getDoc() is unsupported! Use RLCSA instead." << std::endl;
            std::exit(2);
            return 0;
        }
        

    protected:
        // Protected constructor; use TextCollectionBuilder
        TextCollection() { };

//        index_mode_t indexMode;

        // No copy constructor or assignment
        TextCollection(TextCollection const&);
        TextCollection& operator = (TextCollection const&);
    };
}
#endif
