/******************************************************************************
 *   Copyright (C) 2009 by Niko Valimaki <nvalimak@cs.helsinki.fi>            *
 *   Text collection interface for an in-memory XQuery/XPath engine           *
 *                                                                            *
 *   This program is free software; you can redistribute it and/or modify     *
 *   it under the terms of the GNU Lesser General Public License as published *
 *   by the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                      *
 *                                                                            *
 *   This program is distributed in the hope that it will be useful,          *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
 *   GNU Lesser General Public License for more details.                      *
 *                                                                            *
 *   You should have received a copy of the GNU Lesser General Public License *
 *   along with this program; if not, write to the                            *
 *   Free Software Foundation, Inc.,                                          *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                *
 ******************************************************************************/ 

#ifndef _SXSI_TextCollectionBuilder_h_
#define _SXSI_TextCollectionBuilder_h_

#include "TextCollection.h"
#include "TextStorage.h"
#include "Tools.h"

// Default samplerate for suffix array samples
#define TEXTCOLLECTION_DEFAULT_SAMPLERATE 32

// Default input length, used to calculate the buffer size.
#define TEXTCOLLECTION_DEFAULT_INPUT_LENGTH (150 * 1024 * 1024)


namespace SXSI
{
    /**
     * Builder for an instance of the TextCollection class.
     */
    class TextCollectionBuilder
    {
    public:
        // Index type defaults to FM-index.
        // SWCSA can be used for natural language inputs.
        // NB: Current SWCSA uses a lot of memory during construction!
        enum index_type_t { index_type_default, index_type_swcsa, index_type_rlcsa }; 

        static TextCollectionBuilder* create(unsigned samplerate = TEXTCOLLECTION_DEFAULT_SAMPLERATE, 
                                             index_type_t type = index_type_default,
                                             ulong estimatedInputLength =  TEXTCOLLECTION_DEFAULT_INPUT_LENGTH);
        

        virtual ~TextCollectionBuilder() { };
        
        /** 
         * Insert text
         *
         * Must be a zero-terminated string from alphabet [1,255].
         * Can not be called after makeStatic().
         * The i'th text insertion gets an identifier value i-1.
         * In other words, document identifiers start from 0.
         *
         * Second parameter tells if the text will be added to the
         * index also. If false, text is added only to the TextCollection
         * and can not be searched for.
         */
        virtual void InsertText(uchar const *, bool index = true) = 0;
        /**
         * Make static
         *
         * Convert to a static collection.
         * New texts can not be inserted after this operation.
         *
         * TextStorage type defaults to TYPE_PLAIN_TEXT, another
         * possible type is TYPE_LZ_INDEX.
         */
        virtual TextCollection * InitTextCollection(char type = TextStorage::TYPE_PLAIN_TEXT) = 0;
        
    protected:
        // Protected constructor; use the static method TextCollectionBuilder::create()
        TextCollectionBuilder() { };

    private:
        // No copy constructor or assignment
        TextCollectionBuilder(TextCollectionBuilder const&);
        TextCollectionBuilder& operator = (TextCollectionBuilder const&);
    };

}

#endif
