
/* MemoryManager.h
 * Copyright (C) 2011, Antonio Fariña, all rights reserved.
 *
 * MemoryManager.h: reduces the cost of allocating memory when many (millions?)
 *   small pieces of data are allocated. It allocates large-blocks of data as 
 *   needed and returns pointers to them until their are exhausted. Then new
 *   large-blocks are allocated (1 at a time).
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

/*-----------------------------------------------------------------------
 File       : MemoryManager.h
 Function   : Reserves large blocks of memory and gives pointers to small
              portions of that block when requested.
              This improves performance since a unique "LARGE ALLOCATION"
              of memory is needed (a unique call to malloc).
              It is also responsible of freeing memory.
 Last change: 10/03/2004
 Purpose    : Improve hash performance.
 ------------------------------------------------------------------------*/

#ifndef MEMORYMANAGERINCLUDED
#define  MEMORYMANAGERINCLUDED      // only used for hashTable of stopwords
	
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <stdio.h>

#ifndef byte
	#define byte unsigned char
#endif 

#define LARGE_BLOCK_SIZE 1024*256 			// Size of the blocks of memory that will be allocated
#define MAX_BLOCKS 2048						// Maximum number of blocks of size LARGE_BLOCK_SIZE that
											// can be allocated

	/*
	 * Definition of structure MemoryManager
	 */
	struct sMem {
		byte *BLOCKS[MAX_BLOCKS]; 			//array of blocks of size LARGE_BLOCK_SIZE
		unsigned int  currentBlock;			//currentBlock in the array of blocks
		unsigned long remainderBytes; 		//number of bytes not yet assigned in BLOCKS[currentBlock]
		byte *availableByte;				//pointer to next byte not yet assigned
	} ;

	typedef struct sMem *MemoryManager;
	

	MemoryManager createMemoryManager(void);
	void destroyMemoryManager (MemoryManager mm);
	void getMemoryBlock (MemoryManager mm, byte **dst, const unsigned int size);
	void createNewMemoryBlock (MemoryManager mm);

#endif
