#ifndef XML_TREE_INTERNAL__
#error "xml-tree-inc.hpp should not be included directly"
#endif

#ifndef XML_TREE_INC_HPP_
#define XML_TREE_INC_HPP_

#include <cstdio>

inline uint32_t xml_tree::size() const
{
  return tag_seq_len / 2;
}

inline uint32_t xml_tree::num_tags() const
{
  return tag_names->size();
}

inline uint32_t xml_tree::subtree_size(xml_tree::node_t x) const
{
  return x == root() ? this->par->n/2 : bp_subtree_size(this->par, x);
}

inline uint32_t
xml_tree::subtree_tags(xml_tree::node_t x, xml_tree::tag_t label) const
{
  xml_tree::node_t y = bp_find_close(this->par, x);
  if (y - x < 10) {
    uint32_t count = 0;
    for(xml_tree::node_t i = x; i < y; i++)
      count += (tag(i) == label);
    return count;
  } else {
    return tags[label]->rank(y) - tags[label]->rank(x);
  };
}

inline uint32_t xml_tree::subtree_elements(xml_tree::node_t x,
					   xml_tree::tag_t *atts) const
{

  int32_t size = bp_subtree_size(par, x) - 1;
  if (size <= 0) return 0;
  size -= subtree_tags(x, xml_tree::PCDATA_OPEN_TAG_ID);
  if (size < 3) return (uint32_t) size;
  for(; *atts != xml_tree::NIL_TAG_ID; atts++)
    size -= subtree_tags(x, *atts);
  return (uint32_t) size;

}

inline bool xml_tree::is_leaf(xml_tree::node_t x) const
{
  return !bp_inspect(this->par, x+1);
}

inline bool xml_tree::is_open(xml_tree::node_t x) const
{
  return bp_inspect(this->par, x);
}

inline bool xml_tree::is_ancestor(xml_tree::node_t x,
                                  xml_tree::node_t y) const
{
  return bp_is_ancestor(this->par, x, y);
}

inline bool
xml_tree::is_right_descendant(xml_tree::node_t x,
                              xml_tree::node_t y) const
{
  return
    (x > root())
    && (y <= bp_parent_close(this->par, x))
    && (y >= bp_find_close(this->par, x));
}

inline bool xml_tree::is_first_child(xml_tree::node_t x) const
{
  return (x <= this->root()) || (bp_inspect(this->par, x - 1));
}

inline bool xml_tree::is_nil(xml_tree::node_t x) const
{
  return (x == xml_tree::NIL);
}

inline xml_tree::tag_t xml_tree::tag(xml_tree::node_t x) const
{
  if (bits_per_tag == 8)
    return  (xml_tree::tag_t) (((unsigned char*) tag_seq)[x]);
  else
    return get_field(tag_seq, bits_per_tag, x);
}

inline xml_tree::node_t xml_tree::root() const
{
  return xml_tree::ROOT;
}


inline xml_tree::node_t xml_tree::parent(xml_tree::node_t x) const
{
  return bp_parent(this->par, x);
}

inline xml_tree::node_t xml_tree::first_child(node_t x) const
{
  xml_tree::node_t result = bp_first_child(this->par, x);
  return result;
}

inline xml_tree::node_t xml_tree::last_child(xml_tree::node_t x) const
{
  if (is_leaf(x))
    return  xml_tree::NIL;
  else
    return bp_find_open(this->par, bp_find_close(this->par, x) - 1);
}

inline xml_tree::node_t xml_tree::next_sibling(xml_tree::node_t x) const
{
  xml_tree::node_t result = bp_next_sibling(this->par, x);
  return result;
}

inline xml_tree::node_t xml_tree::prev_sibling(xml_tree::node_t x) const
{
  return bp_prev_sibling(this->par, x);
}

inline xml_tree::node_t xml_tree::first_element(xml_tree::node_t x) const
{
  xml_tree::node_t n = first_child(x);
  if (is_nil(n)) return xml_tree::NIL;
  switch (tag(n)) {
  case xml_tree::ATTRIBUTE_OPEN_TAG_ID:
    n = next_sibling(n);
    if (is_nil(n) || tag(n) != xml_tree::PCDATA_OPEN_TAG_ID) return n;
    //Fallthrough
  case PCDATA_OPEN_TAG_ID:
    n = n + 2;
    return bp_inspect(this->par, n) ? n : xml_tree::NIL;
  default:
    return n;
  };
}

inline xml_tree::node_t xml_tree::next_element(xml_tree::node_t x) const
{
  x = next_sibling(x);
  if (is_nil(x)) return x;
  if (tag(x) == xml_tree::PCDATA_OPEN_TAG_ID){
    xml_tree::node_t y = x + 2;
    return bp_inspect(this->par, y) ? y : xml_tree::NIL;
  } else
    return x;
}


inline xml_tree::node_t xml_tree::tagged_next_close(xml_tree::node_t x,
					      xml_tree::tag_t label) const
{
  xml_tree::node_t i=x;
  for(xml_tree::node_t i = x+1; i < std::min(x+100, this->par->n); i++)
    if (tag(i) == label)
      return i;

  return tags[label]->select_next(i);
}

inline xml_tree::node_t xml_tree::tagged_next(xml_tree::node_t x,
					      xml_tree::tag_t label) const
{
  return tags[label]->select_next(x);
}

inline xml_tree::node_t
xml_tree::tagged_descendant(xml_tree::node_t x,
                            xml_tree::tag_t tag) const
{
  xml_tree::node_t y = tagged_next(x, tag);
  return is_nil(y) || !is_ancestor(x, y) ? xml_tree::NIL : y;
}

inline xml_tree::node_t
xml_tree::tagged_following_before(xml_tree::node_t x,
                                  xml_tree::tag_t tag,
				  xml_tree::node_t limit) const
{
  xml_tree::node_t close = bp_find_close(this->par, x);
  xml_tree::node_t s = tagged_next(close, tag);
  return (s < limit) ? s : xml_tree::NIL;
}


inline xml_tree::node_t xml_tree::tagged_child(xml_tree::node_t x,
                                               xml_tree::tag_t t) const
{
  xml_tree::node_t c = first_child(x);
  xml_tree::node_t result;
  if (is_nil(c) || tag(c) == t)
    return c;
  else
    return tagged_sibling(c, t);
}

inline xml_tree::node_t xml_tree::tagged_sibling(xml_tree::node_t x,
                                                 xml_tree::tag_t t) const
{
  xml_tree::node_t sibling = next_sibling(x);
  xml_tree::tag_t stag;
  while (sibling  != xml_tree::NIL) {
    stag = tag(sibling);
    if (stag == t)
      return  sibling;
    sibling = next_sibling(sibling);
  };
  return sibling;
}

inline xml_tree::node_t
xml_tree::select_descendant(xml_tree::node_t x, xml_tree::tag_t *tags) const
{
  if (is_leaf(x)) return xml_tree::NIL;
  auto min = xml_tree::NIL;
  auto aux = xml_tree::NIL;
  for(; *tags != xml_tree::NIL_TAG_ID; ++tags){
    aux = tagged_next(x, *tags);
    if ((unsigned int) aux < (unsigned int) min) min = aux;
  };
  return (min == xml_tree::NIL || is_ancestor(x, min)) ? min : xml_tree::NIL;
}


inline xml_tree::node_t
xml_tree::select_following_before(xml_tree::node_t x, xml_tree::tag_t *tags,
                                  xml_tree::node_t limit) const
{
  auto min = xml_tree::NIL;
  auto aux = xml_tree::NIL;
  auto close = bp_find_close(this->par, x);

  for(; *tags != xml_tree::NIL_TAG_ID; ++tags){
    aux = tagged_next(close, *tags);
    if ((unsigned int) aux < (unsigned int) min) min = aux;
  }

  return (min == xml_tree::NIL || min < limit) ? min : xml_tree::NIL;
}


xml_tree::node_t xml_tree::closing(xml_tree::node_t x) const
{
  return bp_find_close(this->par, x);
}


inline SXSI::TextCollection *xml_tree::get_text_collection() const
{
  return text_collection;
}

inline xml_tree::node_t xml_tree::parent_node(int32_t d) const
{
  xml_tree::node_t res = text_positions->select1(d+1);
  return (xml_tree::node_t) res;
}

inline SXSI::TextCollection::document_result
xml_tree::prefix(uchar const *s) const
{
  return text_collection->Prefix(s);
}

inline SXSI::TextCollection::document_result
xml_tree::suffix(uchar const *s) const
{
  return text_collection->Suffix(s);
}

inline SXSI::TextCollection::document_result
xml_tree::equals(uchar const *s) const
{
  return text_collection->Equal(s);
}

inline SXSI::TextCollection::document_result
xml_tree::contains(uchar const *s) const
{
  return text_collection->Contains(s);
}

inline SXSI::TextCollection::document_result
xml_tree::less_than(uchar const *s) const
{
  return text_collection->LessThan(s);
}


#endif //XML_TREE_INC_HPP_
