extern "C" {
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
}

#include <stdexcept>
#include <cerrno>
#include <cstring>
#include <cstdio>
#include "xml-tree.hpp"

using namespace SXSI;

const xml_tree::node_t xml_tree::NIL;
const xml_tree::node_t xml_tree::ROOT;


const xml_tree::tag_t xml_tree::NIL_TAG_ID;
const char* xml_tree::NIL_TAG = "<!NIL!>";
const xml_tree::tag_t xml_tree::DOCUMENT_OPEN_TAG_ID;
const char* xml_tree::DOCUMENT_OPEN_TAG = "";
const xml_tree::tag_t xml_tree::ATTRIBUTE_OPEN_TAG_ID;
const char* xml_tree::ATTRIBUTE_OPEN_TAG = "<@>";
const xml_tree::tag_t xml_tree::PCDATA_OPEN_TAG_ID;
const char* xml_tree::PCDATA_OPEN_TAG = "<$>";
const xml_tree::tag_t xml_tree::ATTRIBUTE_DATA_OPEN_TAG_ID;
const char* xml_tree::ATTRIBUTE_DATA_OPEN_TAG = "<@$>";
const xml_tree::tag_t xml_tree::CLOSE_TAG_ID;
const char* xml_tree::CLOSE_TAG = "</>";


static int bits8 (int t ) {
  int r = bits(t);
  if (r <= 8)
    return 8;
  else if (r <= 16)
    return 16;
  else
    return r;
}


static bool array_mem(xml_tree::tag_t *a, xml_tree::tag_t t)
{
  for(; *a != xml_tree::NIL_TAG_ID; a++){
    if (*a >= t) return (*a == t);
  };
  return false;
}

static void ufread(void *ptr, size_t size, size_t nmemb, FILE *stream)
 {
    size_t res;
    res = fread(ptr,size,nmemb,stream);
    if (res < nmemb)
      throw std::runtime_error("ufread I/O error" );
    return;
 }

static void ufwrite(const void *ptr, size_t size, size_t nmemb, FILE *stream)
 {
    size_t res;
    res = fwrite(ptr,size,nmemb,stream);
    if (res < nmemb)
      throw std::runtime_error("ufwrite I/O error");
    return;
 }



xml_tree::xml_tree()
{
  print_stack = 0;
  print_buffer = 0;
}

xml_tree::xml_tree(std::vector<int32_t> *tags_,
                   std::unordered_map<std::string, int32_t> *tag_ids,
                   bit_vector *parbitmap,
                   bool disable_tc,
                   SXSI::TextCollectionBuilder *tc_builder,
                   SXSI::TextCollectionBuilder::index_type_t idx_type,
                   bit_vector *textbitmap)
{
  print_stack = 0;
  print_buffer = 0;

  size_t npar = parbitmap->size();
  parbitmap->pack();
  par = bp_construct(npar,
                     parbitmap->get_vector_ptr(),
                     OPT_DEGREE);
  delete parbitmap;

  this->tag_ids = tag_ids;
  tag_names = new std::vector<std::string>();
  tag_names->resize(tag_ids->size());
  for(auto val : *(this->tag_ids))
    (*this->tag_names)[val.second] = val.first;

  uint32_t max_tag = tag_names->size() - 1;

  bit_vector *tmp_bitmap = new bit_vector(npar, 1, 0);
  uint32_t * buff = tmp_bitmap->get_vector_ptr();
  for(int32_t i = 0; i <= max_tag; i++){
    uint32_t ones = 0;
    for (size_t k = 0; k < npar; k++){
      bool t = (i == (*tags_)[k]);
      tmp_bitmap->set(k, t);
      ones += t;
    };

    this->tags.push_back(new static_bitsequence_sdarray(buff, npar, ones));
  }
  delete tmp_bitmap;
  delete [] buff;

  //static_bitsequence_builder *bmb = new static_bitsequence_builder_sdarray();
  //alphabet_mapper *am = new alphabet_mapper_none();

//  this->tags = new static_sequence_bs((uint32_t*)&((*tags)[0]), npar, am, bmb);

  bits_per_tag = bits8(max_tag);
  tag_seq_len = npar;
  tag_seq = new uint32_t[uint_len(bits_per_tag, tag_seq_len)];
  for(size_t i = 0; i < (size_t) tag_seq_len; i++)
    set_field(tag_seq, bits_per_tag, i, (uint32_t) (*tags_)[i]);

  delete tags_;

  if (disable_tc) {
    text_positions = 0;
    text_collection = 0;
  } else {
    textbitmap->pack();
    uint32_t * textbm = textbitmap->get_vector_ptr();
    text_positions = new static_bitsequence_rrr02(textbm,
                                                  npar,
                                                  32);
    //delete [] textbm;
    //delete textbitmap;

    this->text_index_type = idx_type;
    text_collection = tc_builder->InitTextCollection();
    delete tc_builder;
  };

}

xml_tree::~xml_tree()
{
  bp_delete(par);

  for(int i = 0; i < tag_names->size(); i++){
    delete tags[i];
    tags[i] = 0;
  };


  delete [] tag_seq;
  delete tag_names;
  delete tag_ids;
  if (text_collection) delete text_collection;
  if (text_positions) delete text_positions;
}

bool xml_tree::is_child(xml_tree::node_t x, xml_tree::node_t y) const
{
  return !is_leaf(x) && is_ancestor(y, x) && depth(x)+1 == depth(y);
}

uint32_t xml_tree::depth(xml_tree::node_t x) const
{
  return bp_depth(this->par, x);
}


uint32_t xml_tree::preorder(xml_tree::node_t x) const
{
  return bp_preorder_rank(this->par, x);
}

uint32_t xml_tree::postorder(xml_tree::node_t x) const
{
  return bp_postorder_rank(this->par, x);
}

xml_tree::node_t
xml_tree::select_child(xml_tree::node_t x, xml_tree::tag_t *tags) const
{
  if (is_nil(x) || is_leaf(x)) return xml_tree::NIL;
  xml_tree::node_t child = first_child(x);
  if (array_mem(tags, tag(child))) return child;
  return select_sibling(child, tags);
}


xml_tree::node_t
xml_tree::select_sibling(xml_tree::node_t x, xml_tree::tag_t *tags) const
{
  xml_tree::node_t sibling = next_sibling(x);
  xml_tree::tag_t t;
  while(!is_nil(sibling)) {
    t = tag(sibling);
    if (array_mem(tags, t)) return sibling;
    sibling = next_sibling(sibling);
  };
  return sibling;
}


void xml_tree::save(int fd, char* s)
{
  FILE* fp;
  int fd2 = dup(fd);

  fp = fdopen(fd2, "w");

  if (fd == 0) throw(std::runtime_error(strerror(errno)));
  saveTree(par, fp); //TODO use new api

  int ntags = tag_names->size();

  ufwrite(&ntags, sizeof(int), 1, fp);
  for (int i = 0; i < ntags;i++)
    fprintf(fp, "%s\n", tag_names->at(i).c_str());

  //tags->save(fp);
  for(int i = 0; i < ntags; i++)
    tags[i]->save(fp);

  ufwrite(&bits_per_tag, sizeof(uint),1,fp);
  ufwrite(&tag_seq_len, sizeof(uint),1,fp);
  ufwrite(tag_seq, sizeof(uint), uint_len(bits_per_tag, tag_seq_len), fp);

  bool disable_tc = text_collection == 0 || text_positions == 0;

  ufwrite(&disable_tc, sizeof(bool),1,fp);

  if (!disable_tc) {
    text_positions->save(fp);
    ufwrite(&text_index_type,
            sizeof(TextCollectionBuilder::index_type_t),
            1, fp);

    std::string file(s);
    switch (text_index_type){
    case TextCollectionBuilder::index_type_default:
      file.append("_default");
      break;
    case TextCollectionBuilder::index_type_swcsa:
      file.append("_swcsa");
      break;
    case TextCollectionBuilder::index_type_rlcsa:
      file.append("_rlcsa");
      break;
    };

    text_collection->Save(fp, file.c_str());
  };

  fflush(fp);
  fclose(fp);

}

xml_tree* xml_tree::load(int fd, char* name, bool load_tc, int sf)
{
  FILE *fp;
  char buffer[1024];

  int i;
  buffer[1023] = '\0';
  fp = fdopen(fd, "r");
  xml_tree *tree = new xml_tree();

  tree->par = loadTree(fp); //TODO use new api
  tree->tag_names = new std::vector<std::string>();
  tree->tag_ids = new std::unordered_map<std::string, xml_tree::tag_t>();
  std::string s;
  int ntags;

  ufread(&ntags, sizeof(int), 1, fp);

  for (i = 0; i < ntags; i++) {
    if (fgets(buffer, 1022, fp) != buffer)
      throw std::runtime_error("xml_tree::load: cannot read tag list");
    s = buffer;
    // remove the trailing \n
    s.erase(s.size()-1);
    tree->tag_names->push_back(s);
    tree->tag_ids->insert(std::make_pair(s,
                                         static_cast<xml_tree::tag_t>(i)));

  };

  for(int i = 0; i < ntags; i++){
    tree->tags.push_back(static_bitsequence_sdarray::load(fp));
  };

  //tree->tags = static_sequence_bs::load(fp);
  ufread(&tree->bits_per_tag, sizeof(uint), 1, fp);
  fprintf(stderr, "\nBits per tag: %u\n", tree->bits_per_tag);
  ufread(&tree->tag_seq_len, sizeof(uint), 1, fp);
  size_t size = uint_len(tree->bits_per_tag, tree->tag_seq_len);
  tree->tag_seq = new uint[size];
  ufread(tree->tag_seq, sizeof(uint), size, fp);

  bool disable_tc;
  ufread(&disable_tc, sizeof(bool), 1, fp);

  if (disable_tc || !load_tc) {
    tree->text_positions = 0;
    tree->text_collection = 0;
  } else {

    tree->text_positions = static_bitsequence_rrr02::load(fp);

    ufread(&tree->text_index_type,
           sizeof(TextCollectionBuilder::index_type_t), 1, fp);

    std::string file(name);
    switch (tree->text_index_type){
    case TextCollectionBuilder::index_type_default:
      file.append("_default");
      break;
    case TextCollectionBuilder::index_type_swcsa:
      file.append("_swcsa");
      break;
    case TextCollectionBuilder::index_type_rlcsa:
      file.append("_rlcsa");
      break;
    };


    tree->text_collection =
      TextCollection::Load(fp,
                           file.c_str(),
                           TextCollection::index_mode_default,
                           sf);

  };
  return tree;
}



uint32_t xml_tree::num_children(xml_tree::node_t x) const
{
  return bp_degree(par, x);
}

uint32_t xml_tree::child_pos(xml_tree::node_t x) const
{
  return bp_child_rank(par, x);
}

xml_tree::node_t xml_tree::child(xml_tree::node_t x, uint32_t i) const
{
  if (i < 10) return bp_naive_child(par, x, i);
  else bp_child(par, x, i);
}

std::pair<int32_t, int32_t> xml_tree::text_id_range(xml_tree::node_t x) const
{
  int32_t i, j;
  xml_tree::node_t y = bp_find_close(par, x);
  if (x == root())
    i = 0;
  else
  i = text_positions->rank1(x);
  j = text_positions->rank1(y);
//  fprintf(stderr, "Rank of node %i is %i, rank of closing %i is %i\n", x, i, y, j);
  if (i == j)
    return std::make_pair(xml_tree::NIL, xml_tree::NIL);
  else
    return std::make_pair(i, j);
}

int32_t xml_tree::text_id(xml_tree::node_t x) const
{
  return (int32_t) text_positions->rank1(x) - 1;
}

unsigned char* xml_tree::get_text(int32_t id) const
{
  unsigned char * s = text_collection->GetText(id);
  return s + (s[0] == 1);
}

void xml_tree::uflush(int fd)
{
  size_t size = print_buffer->size();
  if (size < BUFFER_SIZE) return;
  uflush_r(fd, size);
}

void xml_tree::flush(int fd)
{
  uflush_r(fd, print_buffer->size());
}


void xml_tree::uflush_r(int fd, size_t s)
{
  if (s == 0 || print_buffer == 0 || fd <= 0) return;
  size_t written;
  while (1) {
    written = write(fd, print_buffer->data(), s);
    if ((written < 0) && (errno == EAGAIN || errno == EINTR)){
      continue;
    };
    break;
  };
  print_buffer->clear();
}

void xml_tree::uput_str(std::string s, int fd)
{
  print_buffer->append(s);
  uflush(fd);
}
void xml_tree::uputs(const char* s, int fd)
{
  print_buffer->append(s);
  uflush(fd);
}
void xml_tree::uputc(const char c, int fd)
{
  print_buffer->push_back(c);
  uflush(fd);
}

const char * xml_tree::get_tag_name_by_ref(xml_tree::tag_t tagid) const
{

   unsigned char *s;
   if (tagid < 0 || tagid >= tag_names->size())
     return "<INVALID TAG>";

   return (const char *) (*tag_names)[tagid].c_str();
}

xml_tree::tag_t xml_tree::register_tag(char *s)
{
  auto found = tag_ids->find(std::string(s));
  if (found == tag_ids->end())
    return xml_tree::NIL_TAG_ID;
  else
    return (*found).second;
}

size_t xml_tree::uprintf(const char*s, int fd)
{
     if (s == 0) return 0;
     size_t i;
     char c;
     for (i = 0; (c = s[i]); i++) {
       switch (c) {
       case '"':
         uputs("&quot;", fd);
         break;
       case '&':
         uputs("&amp;", fd);
         break;
       case '\'':
         uputs("&apos;", fd);
         break;
       case '<':
         uputs("&lt;", fd);
         break;
       case '>':
         uputs("&gt;", fd);
         break;
       default:
         uputc(c, fd);

       };
     };
     return i;
}

void xml_tree::print(xml_tree::node_t x, int fd, bool no_text)
{

  if (print_buffer == 0) {
    print_buffer = new std::string(BUFFER_SIZE, 0);
    print_buffer->clear();
    print_stack = new std::vector<std::string>();
    print_stack->reserve(256);
  };

  xml_tree::node_t fin = bp_find_close(par, x);
  xml_tree::node_t n = x;
  xml_tree::tag_t label = tag(n);
  unsigned char * orig_text;
  unsigned char * current_text;

  auto r = text_id_range(x);
  if (r.first == xml_tree::NIL)
    current_text = 0;
  else {
    current_text = text_collection->GetText(r.first, r.second);
    current_text += (current_text[0] == 1);
  };
//  fprintf(stderr, "Printing node %i, tag: %s\n", x, get_tag_name_by_ref(tag(x)));
//  fprintf(stderr, "Beggining of text is :%s\n", current_text);
//  fflush(stderr);



  orig_text = current_text;

  size_t read = 0;

  while (n <= fin) {

    if (bp_inspect(par, n)) {
      if (label == xml_tree::PCDATA_OPEN_TAG_ID){
        if (no_text) {
          uputs("<$/>", fd);
        } else {
          read = uprintf( (const char*) current_text, fd);
          current_text += read + 1;
        };
        n += 2; // skip closin $
        label = tag(n);
      } else {

        uputc('<', fd);
        uput_str((*tag_names)[label], fd);
        n++;
        if (bp_inspect(par, n)) {
          print_stack->push_back((*tag_names)[label]);
          label = tag(n);
          if (label == xml_tree::ATTRIBUTE_OPEN_TAG_ID) {
            n++;
            if (no_text) uputs("><@@>", fd);

            while (bp_inspect(par, n))
              if (no_text) {
                uputc('<', fd);
                uputs((const char*) &(get_tag_name_by_ref(tag(n))[3]), fd);
                uputc('>', fd);
                uputs("<$@/></", fd);
                uputs((const char*) &(get_tag_name_by_ref(tag(n))[3]), fd);
                uputc('>', fd);
                n += 4;
              } else {
                uputc(' ', fd);
                uputs((const char*) &(get_tag_name_by_ref(tag(n))[3]), fd);
                n++;
                uputs("=\"", fd);
                read = uprintf((const char*) current_text, fd);
                current_text += read + 1;
                uputc('"', fd);
                n += 3;
              };

            if (no_text)
              uputs("</@@>", fd);
            else uputc('>', fd);
            n++;
            label = tag(n);
          } else
            uputc('>', fd);
        } else {
          uputs("/>", fd);
          n++;
          label = tag(n);
        };
      };
    } else do {
      uputs("</", fd);
      uput_str(print_stack->back(), fd);
      uputc('>', fd);
      print_stack->pop_back();
      n++;
      } while (!bp_inspect(par, n) && !print_stack->empty());
    label = tag(n);
  };
  uputc('\n', fd);
  if (orig_text && text_index_type != TextCollectionBuilder::index_type_default)
    if (*orig_text == '\0')
      text_collection->DeleteText(orig_text - 1);
    else
      text_collection->DeleteText(orig_text);

}

// void xml_tree::print(xml_tree::node_t x, int fd, bool no_text)
// {
//   if (print_buffer == 0) {
//     print_buffer = new std::string(BUFFER_SIZE, 0);
//     print_buffer->clear();
//     print_stack = new std::vector<std::string>();
//     print_stack->reserve(256);
//   };

//   xml_tree::node_t fin = bp_find_close(par, x);
//   xml_tree::node_t n = x;
//   xml_tree::tag_t label = tag(n);
//   unsigned char * current_text;


//   while (n <= fin) {

//     if (bp_inspect(par, n)) {
//       if (label == xml_tree::PCDATA_OPEN_TAG_ID){
//         if (no_text) {
//           uputs("<$/>", fd);
//         } else {
// 	  current_text = get_text(text_id(n));
// 	  uprintf( (const char*) (current_text + (current_text[0] == 1)), fd);

// 	  if (current_text && text_index_type != TextCollectionBuilder::index_type_default)
// 	    text_collection->DeleteText(current_text);

// 	  n += 2; // skip closin $
// 	  label = tag(n);
// 	};
//       } else {
// 	  uputc('<', fd);
// 	  uput_str((*tag_names)[label], fd);
// 	  n++;
// 	  if (bp_inspect(par, n)) {
// 	    print_stack->push_back((*tag_names)[label]);
// 	    label = tag(n);
// 	    if (label == xml_tree::ATTRIBUTE_OPEN_TAG_ID) {
// 	      n++;
// 	      if (no_text) uputs("><@@>", fd);

// 	      while (bp_inspect(par, n))
// 		if (no_text) {
// 		  uputc('<', fd);
// 		  uputs((const char*) &(get_tag_name_by_ref(tag(n))[3]), fd);
// 		  uputc('>', fd);
// 		  uputs("<$@/></", fd);
// 		  uputs((const char*) &(get_tag_name_by_ref(tag(n))[3]), fd);
// 		  uputc('>', fd);
// 		  n += 4;
// 		} else {
// 		  uputc(' ', fd);
// 		  uputs((const char*) &(get_tag_name_by_ref(tag(n))[3]), fd);
// 		  n+= 2;
// 		  uputs("=\"", fd);
// 		  current_text = get_text(text_id(n));
// 		  uprintf((const char*) (current_text + (current_text[0] == 1)), fd);
// 		  if (current_text && text_index_type != TextCollectionBuilder::index_type_default)
// 		      text_collection->DeleteText(current_text);
// 		  uputc('"', fd);
// 		  n += 2;
// 		};

// 	      if (no_text)
// 		uputs("</@@>", fd);
// 	      else uputc('>', fd);
// 	      n++;
// 	      label = tag(n);
// 	    } else
// 	      uputc('>', fd);
// 	  } else {
// 	    uputs("/>", fd);
// 	    n++;
// 	    label = tag(n);
// 	  };
// 	};
//       } else do {
// 	  uputs("</", fd);
// 	  uput_str(print_stack->back(), fd);
// 	  uputc('>', fd);
// 	  print_stack->pop_back();
// 	  n++;
// 	} while (!bp_inspect(par, n) && !print_stack->empty());
//       label = tag(n);
//     };
//   uputc('\n', fd);

// }



static inline uchar * next_char(uchar *s, size_t &numtexts)
{
  uchar c;
  s++;
  c = *s;
  while (c <= 1){
    if (c == 0) {
      numtexts--;
      if (numtexts == 0) return 0;
    };
    s++;
    c = *s;
  };
  return s;
}

static inline bool naive_char_contains(uchar const *s, uchar c, size_t numtexts)
{
  uchar sc;
  while(numtexts != 0){
    sc = *s;
    if (c == sc) return true;
    if (sc == 0) numtexts--;
    s++;
  };
  return false;
}

bool xml_tree::naive_contains(xml_tree::node_t n, uchar const *w) const
{
  if (w[0] == 0)
    return true;
  // fprintf(stderr, "Calling naive_contains on node: %i, with tag: %s\n",
  // 	  n,
  // 	  get_tag_name_by_ref(tag(n)));
  // fflush(stderr);
  std::pair<int32_t, int32_t> range = text_id_range(n);

  //pattern is at least one char
  if (range.first == xml_tree::NIL)
    return false;

  uchar * text = this->text_collection->GetText(range.first, range.second);
  size_t num_texts =
    subtree_tags(n, ATTRIBUTE_DATA_OPEN_TAG_ID) +
    subtree_tags(n, PCDATA_OPEN_TAG_ID);

  if (w[1] == 0)
    return naive_char_contains(text, w[0], num_texts);

  //KMP is overkill
  uchar * s = text;
  uchar * ss;
  uchar  const *ww;
  size_t nnum_texts;
  while (true) {
    // fprintf(stderr, "Current char in s: %c, num_texts is: %lu\n", *s, num_texts);
    // fflush(stderr);
    ss = s;
    ww = w;
    nnum_texts = num_texts;
    while (true){
      // fprintf(stderr, "   Current char in w: %c, num_texts is: %lu\n", *ww, num_texts);
      // fprintf(stderr, "   Current char in s: %c\n", *ss);
      // fflush(stderr);

      if (*ww == 0) return true;
      if (*ww != *ss) break;
      ww++;
      ss = next_char(ss, nnum_texts);
      if (ss == 0) return (*ww == 0);
    };
    // fprintf(stderr, "Not found, returning\n");
    // fflush(stderr);
    // fprintf(stderr, "Current string s is: %s\n", s);
    s = next_char(s, num_texts);
    if (s == 0) return false;
//    fprintf(stderr, "After next_char, string s is: %s\n", s);
//    fflush(stderr);
  };

}
