INCLUDE "utils.ml"
open Format


type 'hcons expr =
  | False | True
  | Or of 'hcons * 'hcons
  | And of 'hcons * 'hcons
  | Atom of ([ `Left | `Right | `Epsilon ] * bool * State.t)
  | Pred of Tree.Predicate.t

type 'hcons node = {
  pos : 'hcons expr;
  mutable neg : 'hcons;
  st : StateSet.t * StateSet.t;
  size: int; (* Todo check if this is needed *)
}

external hash_const_variant : [> ] -> int = "%identity"

module rec Node : Hcons.S
  with type data = Data.t = Hcons.Make (Data)
  and Data : Hashtbl.HashedType  with type t = Node.t node =
  struct
    type t =  Node.t node
    let equal x y = x.size == y.size &&
      match x.pos, y.pos with
	| a,b when a == b -> true
	| Or(xf1, xf2), Or(yf1, yf2)
	| And(xf1, xf2), And(yf1,yf2)  -> (xf1 == yf1) && (xf2 == yf2)
	| Atom(d1, p1, s1), Atom(d2 ,p2 ,s2) -> d1 == d2 && p1 == p2 && s1 == s2
	| Pred(p1), Pred(p2) -> p1 == p2
	| _ -> false

    let hash f =
      match f.pos with
	| False -> 0
	| True -> 1
	| Or (f1, f2) -> HASHINT3(PRIME2, Uid.to_int f1.Node.id, Uid.to_int f2.Node.id)
	| And (f1, f2) -> HASHINT3(PRIME3, Uid.to_int f1.Node.id, Uid.to_int f2.Node.id)
	| Atom(d, p, s) -> HASHINT4(PRIME4, hash_const_variant d,vb p,s)
	| Pred(p) -> HASHINT2(PRIME5, Uid.to_int p.Tree.Predicate.id)
  end

type t = Node.t
let hash x = x.Node.key
let uid x = x.Node.id
let equal = Node.equal
let expr f = f.Node.node.pos
let st f = f.Node.node.st
let size f = f.Node.node.size
let compare f1 f2 = compare f1.Node.id  f2.Node.id
let prio f =
  match expr f with
    | True | False -> 10
    | Pred _ -> 9
    | Atom _ -> 8
    | And _ -> 6
    | Or _ -> 1

let rec print ?(parent=false) ppf f =
  if parent then fprintf ppf "(";
  let _ = match expr f with
    | True -> fprintf ppf "%s" Pretty.top
    | False -> fprintf ppf "%s" Pretty.bottom
    | And(f1,f2) ->
	print ~parent:(prio f > prio f1) ppf f1;
	fprintf ppf " %s "  Pretty.wedge;
	print ~parent:(prio f > prio f2) ppf f2;
    | Or(f1,f2) ->
	(print ppf f1);
	fprintf ppf " %s " Pretty.vee;
	(print ppf f2);
    | Atom(dir, b, s) ->
	let _ = flush_str_formatter() in
	let fmt = str_formatter in
	let a_str, d_str =
	  match  dir with
	    | `Left ->  Pretty.down_arrow, Pretty.subscript 1
	    | `Right -> Pretty.down_arrow, Pretty.subscript 2
	    | `Epsilon -> Pretty.epsilon, ""
	in
	  fprintf fmt "%s%s" a_str d_str;
	  State.print fmt s;
	  let str = flush_str_formatter() in
	    if b then fprintf ppf "%s" str
	    else Pretty.pp_overline ppf str
    | Pred p -> fprintf ppf "P%s" (Pretty.subscript (Uid.to_int p.Tree.Predicate.id))
  in
    if parent then fprintf ppf ")"

let print ppf f =  print ~parent:false ppf f

let is_true f = (expr f) == True
let is_false f = (expr f) == False


let cons pos neg s1 s2 size1 size2 =
  let nnode = Node.make { pos = neg; neg = (Obj.magic 0); st = s2; size = size2 } in
  let pnode = Node.make { pos = pos; neg = nnode ; st = s1; size = size1 } in
    (Node.node nnode).neg <- pnode; (* works because the neg field isn't taken into
				      account for hashing ! *)
    pnode,nnode


let empty_pair = StateSet.empty, StateSet.empty
let true_,false_ = cons True False empty_pair empty_pair 0 0
let atom_ d p s =
  let si = StateSet.singleton s in
  let ss = match d with
    | `Left -> si, StateSet.empty
    | `Right -> StateSet.empty, si
    | `Epsilon -> empty_pair (* TODO CHECK *)
  in fst (cons (Atom(d,p,s)) (Atom(d,not p,s)) ss ss 1 1)

let pred_ p =
  let fneg = !(p.Tree.Predicate.node) in
  let pneg = Tree.Predicate.make (ref (fun t n -> not (fneg t n))) in
  fst (cons (Pred p) (Pred pneg) empty_pair empty_pair 1 1)

let not_ f = f.Node.node.neg

let union_pair (l1,r1) (l2, r2) =
  StateSet.mem_union l1 l2,
  StateSet.mem_union r1 r2

let merge_states f1 f2 =
  let sp =
    union_pair (st f1) (st f2)
  and sn =
    union_pair (st (not_ f1)) (st (not_ f2))
  in
    sp,sn

let order f1 f2 = if uid f1  < uid f2 then f2,f1 else f1,f2

let or_ f1 f2 =
  (* Tautologies: x|x, x|not(x) *)

  if equal f1 f2 then f1
  else if equal f1 (not_ f2) then true_

      (* simplification *)
  else if is_true f1 || is_true f2 then true_
  else if is_false f1 && is_false f2 then false_
  else if is_false f1 then f2
  else if is_false f2 then f1

	      (* commutativity of | *)
  else
    let f1, f2 = order f1 f2 in
    let psize = (size f1) + (size f2) in
    let nsize = (size (not_ f1)) + (size (not_ f2)) in
    let sp, sn = merge_states f1 f2 in
      fst (cons (Or(f1,f2)) (And(not_ f1, not_ f2)) sp sn psize nsize)


let and_ f1 f2 =
  not_ (or_ (not_ f1) (not_ f2))


let of_bool = function true -> true_ | false -> false_

let or_pred f1 f2 =
  match expr f1, expr f2 with
    | Pred p1, Pred p2 ->
	let fp1 = !(p1.Tree.Predicate.node)
	and fp2 = !(p2.Tree.Predicate.node) in
	pred_ (Tree.Predicate.make (ref (fun t n -> (fp1 t n) || (fp2 t n))))
    | _ -> or_ f1 f2

let and_pred f1 f2 =
  match expr f1, expr f2 with
      Pred p1, Pred p2 ->
	let fp1 = !(p1.Tree.Predicate.node)
	and fp2 = !(p2.Tree.Predicate.node) in
	pred_ (Tree.Predicate.make (ref (fun t n -> (fp1 t n) && (fp2 t n))))
    | _ -> and_ f1 f2

let rec has_pred f =
  match expr f with
  | Pred _ -> true
  | Or(f1, f2) | And(f1, f2) -> has_pred f1 || has_pred f2
  | _ -> false


module Infix = struct
  let ( +| ) f1 f2 = or_ f1 f2

  let ( *& ) f1 f2 = and_ f1 f2

  let ( *+ ) d s = atom_ d true s
  let ( *- ) d s = atom_ d false s
end
